class Analysis(af.Analysis):
    def __init__(
        self, lens_data, hyper_model_image_1d, hyper_galaxy_image_1d_path_dict
    ):
        """
        An analysis to fit the noise for a single galaxy image.
        Parameters
        ----------
        lens_data: LensData
            Lens data, including an image and noise
        hyper_model_image_1d: ndarray
            An image produce of the overall system by a model
        hyper_galaxy_image_1d_path_dict: ndarray
            The contribution of one galaxy to the model image
        """

        self.lens_data = lens_data

        self.hyper_galaxy_image_1d_path_dict = hyper_galaxy_image_1d_path_dict

        self.hyper_model_image_1d = hyper_model_image_1d

        #    self.check_for_previously_masked_values(array=self.hyper_model_image_1d)
        #    self.check_for_previously_masked_values(array=self.hyper_galaxy_image_1d)

        self.plot_hyper_galaxy_subplot = af.conf.instance.visualize.get(
            "plots", "plot_hyper_galaxy_subplot", bool
        )

    @staticmethod
    def check_for_previously_masked_values(array):
        if not np.all(array) != 0.0:
            raise exc.PhaseException(
                "When mapping a 2D array to a 1D array using lens data, a value "
                "encountered was 0.0 and therefore masked in a previous phase."
            )

    def visualize(self, instance, image_path, during_analysis):

        if self.plot_hyper_galaxy_subplot:

            hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

            hyper_noise_background = self.hyper_noise_background_for_instance(
                instance=instance
            )

            hyper_model_image_2d = self.lens_data.scaled_array_2d_from_array_1d(
                array_1d=self.hyper_model_image_1d
            )

            for galaxy_path, galaxy in instance.path_instance_tuples_for_class(
                g.Galaxy
            ):

                if galaxy_path in self.hyper_galaxy_image_1d_path_dict:
                    hyper_galaxy_image_2d = self.lens_data.scaled_array_2d_from_array_1d(
                        array_1d=self.hyper_galaxy_image_1d_path_dict[galaxy_path]
                    )

                # contribution_map_2d = galaxy.hyper_galaxy.contribution_map_from_hyper_images(
                #     hyper_model_image=hyper_model_image_2d,
                #     hyper_galaxy_image=hyper_galaxy_image_2d,
                # )

                fit_normal = lens_fit.LensDataFit(
                    image_1d=self.lens_data.image_1d,
                    noise_map_1d=self.lens_data.noise_map_1d,
                    mask_1d=self.lens_data.mask_1d,
                    model_image_1d=self.hyper_model_image_1d,
                    grid_stack=self.lens_data.grid_stack,
                )

                fit = self.fit_for_hyper_galaxy(
                    instance=instance,
                    hyper_galaxy=galaxy.hyper_galaxy,
                    hyper_image_sky=hyper_image_sky,
                    hyper_noise_background=hyper_noise_background,
                )

                hyper_plotters.plot_hyper_galaxy_subplot(
                    hyper_galaxy_image=hyper_galaxy_image_2d,
                    contribution_map=contribution_map_2d,
                    noise_map=self.lens_data.noise_map(return_in_2d=True),
                    hyper_noise_map=fit.noise_map(return_in_2d=True),
                    chi_squared_map=fit_normal.chi_squared_map(return_in_2d=True),
                    hyper_chi_squared_map=fit.chi_squared_map(return_in_2d=True),
                    output_path=image_path,
                    output_format="png",
                )

    def fit(self, instance):
        """
        Fit the model image to the real image by scaling the hyper noise.
        Parameters
        ----------
        instance: ModelInstance
            A model instance with a hyper galaxy property
        Returns
        -------
        fit: float
        """

        hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

        hyper_noise_background = self.hyper_noise_background_for_instance(
            instance=instance
        )

        fit = self.fit_for_hyper_galaxy(
            instance=instance,
            hyper_galaxy=instance.hyper_galaxy,
            hyper_image_sky=hyper_image_sky,
            hyper_noise_background=hyper_noise_background,
        )

        return fit.figure_of_merit

    @staticmethod
    def hyper_image_sky_for_instance(instance):
        if hasattr(instance, "hyper_image_sky"):
            return instance.hyper_image_sky

    @staticmethod
    def hyper_noise_background_for_instance(instance):
        if hasattr(instance, "hyper_noise_background"):
            return instance.hyper_noise_background

    def fit_for_hyper_galaxy(
        self, instance, hyper_galaxy, hyper_image_sky, hyper_noise_background
    ):

        if hyper_image_sky is not None:
            image_1d = hyper_image_sky.image_scaled_sky_from_image(
                image=self.lens_data.image_1d
            )
        else:
            image_1d = self.lens_data.image_1d

        if hyper_noise_background is not None:
            noise_map_1d = hyper_noise_background.noise_map_scaled_noise_from_noise_map(
                noise_map=self.lens_data.noise_map_1d
            )
        else:
            noise_map_1d = self.lens_data.noise_map_1d

        for galaxy_path, galaxy in instance.path_instance_tuples_for_class(g.Galaxy):

            hyper_noise_1d = np.zeros(shape=self.hyper_model_image_1d.shape)

            if galaxy_path in self.hyper_galaxy_image_1d_path_dict:
                hyper_galaxy_image_1d = self.hyper_galaxy_image_1d_path_dict[
                    galaxy_path
                ]

                hyper_noise_1d += hyper_galaxy.hyper_noise_map_from_hyper_images_and_noise_map(
                    hyper_model_image=self.hyper_model_image_1d,
                    hyper_galaxy_image=hyper_galaxy_image_1d,
                    noise_map=self.lens_data.noise_map_1d,
                )

        hyper_noise_map_1d = noise_map_1d + hyper_noise_1d

        return lens_fit.LensDataFit(
            image_1d=image_1d,
            noise_map_1d=hyper_noise_map_1d,
            mask_1d=self.lens_data.mask_1d,
            model_image_1d=self.hyper_model_image_1d,
            grid_stack=self.lens_data.grid_stack,
        )

    @classmethod
    def describe(cls, instance):
        return "Running hyper galaxy fit for HyperGalaxy:\n{}".format(
            instance.hyper_galaxy
        )


def run_hyper(self, data, results=None, mask=None, positions=None):
    """
    Run a fit for each galaxy from the previous phase.
    Parameters
    ----------
    data: LensData
    results: ResultsCollection
        Results from all previous phases
    mask: Mask
        The mask
    positions
    Returns
    -------
    results: HyperGalaxyResults
        A collection of results, with one item per a galaxy
    """
    phase = self.make_hyper_phase()

    mask = setup_phase_mask(
        data=data,
        mask=mask,
        mask_function=cast(phase_imaging.PhaseImaging, phase).mask_function,
        inner_mask_radii=cast(phase_imaging.PhaseImaging, phase).inner_mask_radii,
    )

    lens_data = ld.LensData(
        ccd_data=data,
        mask=mask,
        sub_grid_size=cast(phase_imaging.PhaseImaging, phase).sub_grid_size,
        image_psf_shape=cast(phase_imaging.PhaseImaging, phase).image_psf_shape,
        positions=positions,
        interp_pixel_scale=cast(phase_imaging.PhaseImaging, phase).interp_pixel_scale,
        cluster_pixel_scale=cast(phase_imaging.PhaseImaging, phase).cluster_pixel_scale,
        cluster_pixel_limit=cast(phase_imaging.PhaseImaging, phase).cluster_pixel_limit,
        uses_inversion=cast(phase_imaging.PhaseImaging, phase).uses_inversion,
        uses_cluster_inversion=cast(
            phase_imaging.PhaseImaging, phase
        ).uses_cluster_inversion,
    )

    hyper_model_image_1d = results.last.hyper_model_image_1d_from_mask(
        mask=lens_data.mask_2d
    )
    hyper_galaxy_image_1d_path_dict = results.last.hyper_galaxy_image_1d_path_dict_from_mask(
        mask=lens_data.mask_2d
    )

    hyper_result = copy.deepcopy(results.last)
    hyper_result.variable = hyper_result.variable.copy_with_fixed_priors(
        hyper_result.constant
    )
    hyper_result.analysis.uses_hyper_images = True
    hyper_result.analysis.hyper_model_image_1d = hyper_model_image_1d
    hyper_result.analysis.hyper_galaxy_image_1d_path_dict = (
        hyper_galaxy_image_1d_path_dict
    )

    for path, galaxy in results.last.path_galaxy_tuples:

        optimizer = phase.optimizer.copy_with_name_extension(extension=path[-1])

        optimizer.phase_tag = ""

        # TODO : This is a HACK :O

        optimizer.variable.lens_galaxies = results.last.constant.lens_galaxies

        if hasattr(results.last.constant, "source_galaxies"):
            optimizer.variable.source_galaxies = results.last.constant.source_galaxies

        if hasattr(results.last.constant, "galaxies"):
            optimizer.variable.galaxies = results.last.constant.galaxies

        phase.const_efficiency_mode = af.conf.instance.non_linear.get(
            "MultiNest", "extension_hyper_galaxy_const_efficiency_mode", bool
        )
        phase.optimizer.sampling_efficiency = af.conf.instance.non_linear.get(
            "MultiNest", "extension_hyper_galaxy_sampling_efficiency", float
        )
        phase.optimizer.n_live_points = af.conf.instance.non_linear.get(
            "MultiNest", "extension_hyper_galaxy_n_live_points", int
        )

        optimizer.variable.hyper_galaxy = g.HyperGalaxy

        if self.include_sky_background:
            optimizer.variable.hyper_image_sky = hd.HyperImageSky

        if self.include_noise_background:
            optimizer.variable.hyper_noise_background = hd.HyperNoiseBackground

        # If array is all zeros, galaxy did not have image in previous phase and
        # should be ignored
        if not np.all(hyper_galaxy_image_1d_path_dict[path] == 0):

            analysis = self.Analysis(
                lens_data=lens_data,
                hyper_model_image_1d=hyper_model_image_1d,
                hyper_galaxy_image_1d_path_dict=hyper_galaxy_image_1d_path_dict,
            )

            result = optimizer.fit(analysis)

            def transfer_field(name):
                if hasattr(result.constant, name):
                    setattr(
                        hyper_result.constant.object_for_path(path),
                        name,
                        getattr(result.constant, name),
                    )
                    setattr(
                        hyper_result.variable.object_for_path(path),
                        name,
                        getattr(result.variable, name),
                    )

            transfer_field("hyper_galaxy")
            transfer_field("hyper_image_sky")
            transfer_field("hyper_noise_background")

    return hyper_result



import autofit as af
import sys

from autolens.data import ccd
from autolens.data.array import mask as msk
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

import os

# Given your username is where your data is stored, you'll need to put your cosma username here.
cosma_username = "pdtw24"
cosma_path = "/cosma7/data/dp004/dc-nigh1/autolens/"

data_folder = "subhalo_challenge"

cosma_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=cosma_path, folder_names=["data", data_folder]
)

data_folder = "subhalo_challenge_hyper"

cosma_output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=cosma_path, folder_names=["output", data_folder]
)

workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path, and override the output path with the Cosma path.
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=cosma_output_path
)

# The fifth line of this batch script - '#SBATCH --array=1-17' is what species this. Its telling Cosma we're going to
# run 17 jobs, and the id's of those jobs will be numbered from 1 to 17. Infact, these ids are passed to this runner,
# and we'll use them to ensure that each jobs loads a different image. Lets get the cosma array id for our job.
cosma_array_id = int(sys.argv[1])

### Subhalo challenge data strings ###

data_type = "noise_normal"

if cosma_array_id == 1 or cosma_array_id == 2:
    data_level = "level_0"
else:
    data_level = "level_1"

pixel_scale = 0.00976562

data_name = []
data_name.append("")  # Task number beings at 1, so keep index 0 blank
data_name.append("large_hi_sn_system_1")  # Index 1
data_name.append("small_hi_sn_system_1")  # Index 2

data_name.append("large_hi_sn_system_1")  # Index 3
data_name.append("large_hi_sn_system_2")  # Index 4
data_name.append("large_md_sn_system_1")  # Index 5
data_name.append("large_md_sn_system_2")  # Index 6
data_name.append("large_lo_sn_system_1")  # Index 7
data_name.append("large_lo_sn_system_2")  # Index 8
data_name.append("small_hi_sn_system_1")  # Index 9
data_name.append("small_hi_sn_system_2")  # Index 10
data_name.append("small_md_sn_system_1")  # Index 11
data_name.append("small_md_sn_system_2")  # Index 12
data_name.append("small_lo_sn_system_1")  # Index 13
data_name.append("small_lo_sn_system_2")  # Index 14

data_name = data_name[cosma_array_id]

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=cosma_data_path, folder_names=[data_type, data_level, data_name]
)

ccd_data = ccd.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    pixel_scale=pixel_scale,
    resized_psf_shape=(11, 11),
)

mask = msk.load_mask_from_fits(
    mask_path=data_path + "mask_irregular.fits", pixel_scale=pixel_scale
)

positions = ccd.load_positions(positions_path=data_path + "positions.dat")

from workspace_jam.pipelines.hyper.no_lens_light.initialize import (
    lens_sie_shear_source_sersic,
)
from workspace_jam.pipelines.hyper.no_lens_light.inversion.from_initialize import (
    lens_sie_shear_source_inversion,
)
from workspace_jam.pipelines.hyper.no_lens_light.power_law.from_inversion import (
    lens_pl_shear_source_inversion,
)

# from workspace_jam.pipelines.no_lens_light.subhalo.from_power_law import lens_pl_shear_subhalo_source_inversion

pl_pixelization = pix.VoronoiBrightnessImage
pl_regularization = reg.AdaptiveBrightness

pipeline_initialize = lens_sie_shear_source_sersic.make_pipeline(
    pl_hyper_galaxies=True,
    pl_hyper_background_sky=False,
    pl_hyper_background_noise=False,
    phase_folders=[data_type, data_level, data_name],
    positions_threshold=1.0,
)

pipeline_inversion = lens_sie_shear_source_inversion.make_pipeline(
    pl_hyper_galaxies=True,
    pl_hyper_background_sky=False,
    pl_hyper_background_noise=False,
    pl_pixelization=pl_pixelization,
    pl_regularization=pl_regularization,
    phase_folders=[data_type, data_level, data_name],
    positions_threshold=1.0,
    inversion_pixel_limit=800,
)

pipeline_power_law = lens_pl_shear_source_inversion.make_pipeline(
    pl_hyper_galaxies=True,
    pl_hyper_background_sky=False,
    pl_hyper_background_noise=False,
    pl_pixelization=pl_pixelization,
    pl_regularization=pl_regularization,
    phase_folders=[data_type, data_level, data_name],
    positions_threshold=0.5,
    inversion_pixel_limit=800,
)

# pipeline_subhalo = lens_pl_shear_subhalo_source_inversion.make_pipeline(phase_folders=[data_type, data_level, data_name])

pipeline = (
    pipeline_initialize + pipeline_inversion + pipeline_power_law
)  # + pipeline_subhalo

pipeline.run(data=ccd_data, mask=mask, positions=positions)



import time

from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.lens import lens_data as ld
from autolens.lens import ray_tracing
from autolens.lens import lens_fit
from autolens.lens.util import lens_fit_util

from test.simulation import simulation_util

repeats = 10

print("Number of repeats = " + str(repeats))
print()

sub_grid_size = 4
radius_arcsec = 3.0
psf_shape = (21, 21)

print("sub grid size = " + str(sub_grid_size))
print("circular mask radius = " + str(radius_arcsec) + "\n")
print("psf shape = " + str(psf_shape) + "\n")

lens_galaxy = g.Galaxy(
    light=lp.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.9,
        phi=45.0,
        intensity=0.5,
        effective_radius=0.8,
        sersic_index=4.0,
    ),
    mass=mp.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
)

source_galaxy = g.Galaxy(
    light=lp.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=60.0,
        intensity=0.4,
        effective_radius=0.5,
        sersic_index=1.0,
    )
)

for data_resolution in ["LSST", "Euclid", "HST", "HST_Up", "AO"]:

    ccd_data = simulation_util.load_test_ccd_data(
        data_type="lens_and_source_smooth",
        data_resolution=data_resolution,
        psf_shape=psf_shape,
    )
    mask = msk.Mask.circular(
        shape=ccd_data.shape,
        pixel_scale=ccd_data.pixel_scale,
        radius_arcsec=radius_arcsec,
    )
    lens_data = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=sub_grid_size)

    print("Light profile fit run times for image type " + data_resolution + "\n")
    print("Number of points = " + str(lens_data.grid_stack.sub.shape[0]) + "\n")

    start_overall = time.time()

    start = time.time()
    for i in range(repeats):
        tracer = ray_tracing.TracerImageSourcePlanes(
            lens_galaxies=[lens_galaxy],
            source_galaxies=[source_galaxy],
            image_plane_grid_stack=lens_data.grid_stack,
        )
    diff = time.time() - start
    print("Time to Setup Tracer = {}".format(diff / repeats))

    start = time.time()
    for i in range(repeats):
        blurred_profile_image_1d = lens_fit_util.blurred_image_1d_from_1d_unblurred_and_blurring_images(
            unblurred_image_1d=tracer.profile_image_plane_image,
            blurring_image_1d=tracer.profile_image_plane_blurring_image,
            convolver=lens_data.convolver_image,
        )
        blurred_profile_image = lens_data.scaled_array_2d_from_array_1d(
            array_1d=blurred_profile_image_1d
        )
    diff = time.time() - start
    print("Time to perform PSF convolution = {}".format(diff / repeats))

    start = time.time()
    for i in range(repeats):
        lens_fit.LensDataFit(
            image_1d=lens_data.image_1d,
            noise_map_1d=lens_data.noise_map_1d,
            mask_1d=lens_data.mask_1d,
            model_image_1d=blurred_profile_image_1d,
        )
    diff = time.time() - start
    print("Time to perform fit (1D) = {}".format(diff / repeats))

    start = time.time()
    for i in range(repeats):
        lens_fit.LensDataFit(
            image_1d=lens_data.unmasked_image,
            noise_map_1d=lens_data.unmasked_noise_map,
            mask_1d=lens_data.mask_2d,
            model_image_1d=blurred_profile_image,
        )
    diff = time.time() - start
    print("Time to perform fit (2D) = {}".format(diff / repeats))

    start = time.time()
    for i in range(repeats):
        lens_fit.LensProfileFit(lens_data=lens_data, tracer=tracer)
    diff = time.time() - start
    print("Time to perform complete fit = {}".format(diff / repeats))

    print()



import os

import pytest

from autolens.data.plotters import ccd_plotters


@pytest.fixture(name="ccd_plotter_path")
def make_ccd_plotter_setup():
    ccd_plotter_path = "{}/../../test_files/plotting/ccd/".format(
        os.path.dirname(os.path.realpath(__file__))
    )

    return ccd_plotter_path


def test__individual_attributes_are_output(
    ccd_data_7x7, positions_7x7, mask_7x7, ccd_plotter_path, plot_patch
):
    ccd_plotters.plot_image(
        ccd_data=ccd_data_7x7,
        positions=positions_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ccd_plotter_path,
        output_format="png",
    )

    assert ccd_plotter_path + "ccd_image.png" in plot_patch.paths

    ccd_plotters.plot_noise_map(
        ccd_data=ccd_data_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ccd_plotter_path,
        output_format="png",
    )

    assert ccd_plotter_path + "ccd_noise_map.png" in plot_patch.paths

    ccd_plotters.plot_psf(
        ccd_data=ccd_data_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ccd_plotter_path,
        output_format="png",
    )

    assert ccd_plotter_path + "ccd_psf.png" in plot_patch.paths

    ccd_plotters.plot_signal_to_noise_map(
        ccd_data=ccd_data_7x7,
        mask=mask_7x7,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ccd_plotter_path,
        output_format="png",
    )

    assert ccd_plotter_path + "ccd_signal_to_noise_map.png" in plot_patch.paths

    ccd_plotters.plot_ccd_subplot(
        ccd_data=ccd_data_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ccd_plotter_path,
        output_format="png",
    )

    assert ccd_plotter_path + "ccd_data.png" in plot_patch.paths


def test__ccd_individuals__output_dependent_on_input(
    ccd_data_7x7, general_config, ccd_plotter_path, plot_patch
):
    ccd_plotters.plot_ccd_individual(
        ccd_data=ccd_data_7x7,
        should_plot_image=True,
        should_plot_psf=True,
        should_plot_absolute_signal_to_noise_map=True,
        output_path=ccd_plotter_path,
        output_format="png",
    )

    assert ccd_plotter_path + "ccd_image.png" in plot_patch.paths

    assert not ccd_plotter_path + "ccd_noise_map.png" in plot_patch.paths

    assert ccd_plotter_path + "ccd_psf.png" in plot_patch.paths

    assert not ccd_plotter_path + "ccd_signal_to_noise_map.png" in plot_patch.paths

    assert ccd_plotter_path + "ccd_absolute_signal_to_noise_map.png" in plot_patch.paths

    assert (
        not ccd_plotter_path + "ccd_potential_chi_squared_map.png" in plot_patch.paths
    )



import numpy as np
import pytest

from autolens import exc
from autolens.data.array import grids, mask
from autolens.model.galaxy import galaxy as g
from autolens.model.galaxy.util import galaxy_util
from autolens.lens.util import lens_util
from workspace_jam.scraps.stack import plane_stack as pl_stack
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.unit.mock.model.mock_inversion import MockRegularization, MockPixelization
from test.unit.mock.data.mock_grids import MockBorders


@pytest.fixture(name="grid_stack_0")
def make_grid_stack_0():
    ma = mask.Mask(
        np.array(
            [
                [True, True, True, True],
                [True, False, False, True],
                [True, True, True, True],
            ]
        ),
        pixel_scale=6.0,
    )

    grid_stack_0 = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
        mask=ma, sub_grid_size=2, psf_shape=(3, 3)
    )

    # Manually overwrite a set of cooridnates to make tests of grid_stacks and defledctions straightforward

    grid_stack_0.regular[0] = np.array([1.0, 1.0])
    grid_stack_0.regular[1] = np.array([1.0, 0.0])
    grid_stack_0.sub[0] = np.array([1.0, 1.0])
    grid_stack_0.sub[1] = np.array([1.0, 0.0])
    grid_stack_0.sub[2] = np.array([1.0, 1.0])
    grid_stack_0.sub[3] = np.array([1.0, 0.0])
    grid_stack_0.sub[4] = np.array([-1.0, 2.0])
    grid_stack_0.sub[5] = np.array([-1.0, 4.0])
    grid_stack_0.sub[6] = np.array([1.0, 2.0])
    grid_stack_0.sub[7] = np.array([1.0, 4.0])
    grid_stack_0.blurring[0] = np.array([1.0, 0.0])
    grid_stack_0.blurring[1] = np.array([-6.0, -3.0])
    grid_stack_0.blurring[2] = np.array([-6.0, 3.0])
    grid_stack_0.blurring[3] = np.array([-6.0, 9.0])
    grid_stack_0.blurring[4] = np.array([0.0, -9.0])
    grid_stack_0.blurring[5] = np.array([0.0, 9.0])
    grid_stack_0.blurring[6] = np.array([6.0, -9.0])
    grid_stack_0.blurring[7] = np.array([6.0, -3.0])
    grid_stack_0.blurring[8] = np.array([6.0, 3.0])
    grid_stack_0.blurring[9] = np.array([6.0, 9.0])

    return grid_stack_0


@pytest.fixture(name="grid_stack_1")
def make_grid_stack_1():
    ma = mask.Mask(
        np.array(
            [
                [True, True, True, True],
                [True, False, False, True],
                [True, True, True, True],
            ]
        ),
        pixel_scale=12.0,
    )

    grid_stack_0 = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
        mask=ma, sub_grid_size=2, psf_shape=(3, 3)
    )

    # Manually overwrite a set of cooridnates to make tests of grid_stacks and defledctions straightforward

    grid_stack_0.regular[0] = np.array([2.0, 2.0])
    grid_stack_0.regular[1] = np.array([2.0, 2.0])
    grid_stack_0.sub[0] = np.array([2.0, 2.0])
    grid_stack_0.sub[1] = np.array([2.0, 0.0])
    grid_stack_0.sub[2] = np.array([2.0, 2.0])
    grid_stack_0.sub[3] = np.array([2.0, 0.0])
    grid_stack_0.sub[4] = np.array([-2.0, 4.0])
    grid_stack_0.sub[5] = np.array([-2.0, 8.0])
    grid_stack_0.sub[6] = np.array([2.0, 4.0])
    grid_stack_0.sub[7] = np.array([2.0, 8.0])
    grid_stack_0.blurring[0] = np.array([2.0, 0.0])
    grid_stack_0.blurring[1] = np.array([-12.0, -6.0])
    grid_stack_0.blurring[2] = np.array([-12.0, 6.0])
    grid_stack_0.blurring[3] = np.array([-12.0, 18.0])
    grid_stack_0.blurring[4] = np.array([0.0, -18.0])
    grid_stack_0.blurring[5] = np.array([0.0, 18.0])
    grid_stack_0.blurring[6] = np.array([12.0, -18.0])
    grid_stack_0.blurring[7] = np.array([12.0, -6.0])
    grid_stack_0.blurring[8] = np.array([12.0, 6.0])
    grid_stack_0.blurring[9] = np.array([12.0, 18.0])

    return grid_stack_0


@pytest.fixture(name="padded_grid_stack")
def make_padded_grid_stack():
    ma = mask.Mask(np.array([[True, False]]), pixel_scale=3.0)
    return grids.GridStack.padded_grid_stack_from_psf_shape(ma, 2, (3, 3))


@pytest.fixture(name="galaxy_non", scope="function")
def make_galaxy_non():
    return g.Galaxy()


@pytest.fixture(name="galaxy_light")
def make_galaxy_light():
    return g.Galaxy(
        light_profile=lp.EllipticalSersic(
            centre=(0.1, 0.1),
            axis_ratio=1.0,
            phi=0.0,
            intensity=1.0,
            effective_radius=0.6,
            sersic_index=4.0,
        )
    )


@pytest.fixture(name="galaxy_mass")
def make_galaxy_mass():
    return g.Galaxy(mass_profile=mp.SphericalIsothermal(einstein_radius=1.0))


@pytest.fixture(name="galaxy_mass_x2")
def make_galaxy_mass_x2():
    return g.Galaxy(
        sis_0=mp.SphericalIsothermal(einstein_radius=1.0),
        sis_1=mp.SphericalIsothermal(einstein_radius=1.0),
    )


class TestPlaneStack:
    class TestGridsStacksSetup:
        def test__grid_stack_0_setup_for_regular_sub_and_blurring__no_deflections(
            self, grid_stack_0, galaxy_mass
        ):
            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_mass],
                grid_stacks=[grid_stack_0, grid_stack_0],
                compute_deflections=False,
            )

            assert plane_stack.grid_stacks[0].regular == pytest.approx(
                np.array([[1.0, 1.0], [1.0, 0.0]]), 1e-3
            )
            assert plane_stack.grid_stacks[0].sub == pytest.approx(
                np.array(
                    [
                        [1.0, 1.0],
                        [1.0, 0.0],
                        [1.0, 1.0],
                        [1.0, 0.0],
                        [-1.0, 2.0],
                        [-1.0, 4.0],
                        [1.0, 2.0],
                        [1.0, 4.0],
                    ]
                ),
                1e-3,
            )
            assert plane_stack.grid_stacks[0].blurring == pytest.approx(
                np.array(
                    [
                        [1.0, 0.0],
                        [-6.0, -3.0],
                        [-6.0, 3.0],
                        [-6.0, 9.0],
                        [0.0, -9.0],
                        [0.0, 9.0],
                        [6.0, -9.0],
                        [6.0, -3.0],
                        [6.0, 3.0],
                        [6.0, 9.0],
                    ]
                ),
                1e-3,
            )

            assert plane_stack.grid_stacks[1].regular == pytest.approx(
                np.array([[1.0, 1.0], [1.0, 0.0]]), 1e-3
            )
            assert plane_stack.grid_stacks[1].sub == pytest.approx(
                np.array(
                    [
                        [1.0, 1.0],
                        [1.0, 0.0],
                        [1.0, 1.0],
                        [1.0, 0.0],
                        [-1.0, 2.0],
                        [-1.0, 4.0],
                        [1.0, 2.0],
                        [1.0, 4.0],
                    ]
                ),
                1e-3,
            )
            assert plane_stack.grid_stacks[1].blurring == pytest.approx(
                np.array(
                    [
                        [1.0, 0.0],
                        [-6.0, -3.0],
                        [-6.0, 3.0],
                        [-6.0, 9.0],
                        [0.0, -9.0],
                        [0.0, 9.0],
                        [6.0, -9.0],
                        [6.0, -3.0],
                        [6.0, 3.0],
                        [6.0, 9.0],
                    ]
                ),
                1e-3,
            )

            assert plane_stack.deflection_stacks == None

        def test__same_as_above_but_test_deflections(self, grid_stack_0, galaxy_mass):
            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_mass],
                grid_stacks=[grid_stack_0, grid_stack_0],
                compute_deflections=True,
            )

            sub_galaxy_deflections = galaxy_mass.deflections_from_grid(grid_stack_0.sub)
            blurring_galaxy_deflections = galaxy_mass.deflections_from_grid(
                grid_stack_0.blurring
            )

            assert plane_stack.deflection_stacks[0].regular == pytest.approx(
                np.array([[0.707, 0.707], [1.0, 0.0]]), 1e-3
            )
            assert (
                plane_stack.deflection_stacks[0].sub == sub_galaxy_deflections
            ).all()
            assert (
                plane_stack.deflection_stacks[0].blurring == blurring_galaxy_deflections
            ).all()

            assert plane_stack.deflection_stacks[1].regular == pytest.approx(
                np.array([[0.707, 0.707], [1.0, 0.0]]), 1e-3
            )
            assert (
                plane_stack.deflection_stacks[1].sub == sub_galaxy_deflections
            ).all()
            assert (
                plane_stack.deflection_stacks[1].blurring == blurring_galaxy_deflections
            ).all()

        def test__same_as_above__x2_galaxy_in_plane__or_galaxy_x2_sis__deflections_double(
            self, grid_stack_0, galaxy_mass, galaxy_mass_x2
        ):
            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_mass_x2],
                grid_stacks=[grid_stack_0, grid_stack_0],
                compute_deflections=True,
            )

            sub_galaxy_deflections = galaxy_mass_x2.deflections_from_grid(
                grid_stack_0.sub
            )
            blurring_galaxy_deflections = galaxy_mass_x2.deflections_from_grid(
                grid_stack_0.blurring
            )

            assert plane_stack.deflection_stacks[0].regular == pytest.approx(
                np.array([[2.0 * 0.707, 2.0 * 0.707], [2.0, 0.0]]), 1e-3
            )
            assert (
                plane_stack.deflection_stacks[0].sub == sub_galaxy_deflections
            ).all()
            assert (
                plane_stack.deflection_stacks[0].blurring == blurring_galaxy_deflections
            ).all()

            assert plane_stack.deflection_stacks[1].regular == pytest.approx(
                np.array([[2.0 * 0.707, 2.0 * 0.707], [2.0, 0.0]]), 1e-3
            )
            assert (
                plane_stack.deflection_stacks[1].sub == sub_galaxy_deflections
            ).all()
            assert (
                plane_stack.deflection_stacks[1].blurring == blurring_galaxy_deflections
            ).all()

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_mass, galaxy_mass],
                grid_stacks=[grid_stack_0, grid_stack_0],
                compute_deflections=True,
            )

            sub_galaxy_deflections = galaxy_mass.deflections_from_grid(grid_stack_0.sub)
            blurring_galaxy_deflections = galaxy_mass.deflections_from_grid(
                grid_stack_0.blurring
            )

            assert plane_stack.deflection_stacks[0].regular == pytest.approx(
                np.array([[2.0 * 0.707, 2.0 * 0.707], [2.0, 0.0]]), 1e-3
            )
            assert (
                plane_stack.deflection_stacks[0].sub == 2.0 * sub_galaxy_deflections
            ).all()
            assert (
                plane_stack.deflection_stacks[0].blurring
                == 2.0 * blurring_galaxy_deflections
            ).all()

            assert plane_stack.deflection_stacks[1].regular == pytest.approx(
                np.array([[2.0 * 0.707, 2.0 * 0.707], [2.0, 0.0]]), 1e-3
            )
            assert (
                plane_stack.deflection_stacks[1].sub == 2.0 * sub_galaxy_deflections
            ).all()
            assert (
                plane_stack.deflection_stacks[1].blurring
                == 2.0 * blurring_galaxy_deflections
            ).all()

    class TestProperties:
        def test__total_images(self, grid_stack_0):

            plane_stack = pl_stack.PlaneStack(
                galaxies=[g.Galaxy()], grid_stacks=[grid_stack_0]
            )
            assert plane_stack.total_grid_stacks == 1

            plane_stack = pl_stack.PlaneStack(
                galaxies=[g.Galaxy()], grid_stacks=[grid_stack_0, grid_stack_0]
            )
            assert plane_stack.total_grid_stacks == 2

            plane_stack = pl_stack.PlaneStack(
                galaxies=[g.Galaxy()],
                grid_stacks=[grid_stack_0, grid_stack_0, grid_stack_0],
            )
            assert plane_stack.total_grid_stacks == 3

        def test__padded_grid_in__tracer_has_padded_grid_property(
            self, grid_stack_0, padded_grid_stack, galaxy_light
        ):

            plane_stack = pl_stack.PlaneStack(
                grid_stacks=[grid_stack_0], galaxies=[galaxy_light]
            )
            assert plane_stack.has_padded_grid_stack == False

            plane_stack = pl_stack.PlaneStack(
                grid_stacks=[padded_grid_stack], galaxies=[galaxy_light]
            )
            assert plane_stack.has_padded_grid_stack == True

            plane_stack = pl_stack.PlaneStack(
                grid_stacks=[grid_stack_0, padded_grid_stack], galaxies=[galaxy_light]
            )
            assert plane_stack.has_padded_grid_stack == True

    class TestImages:
        def test__images_from_plane__same_as_light_profile_images(
            self, grid_stack_0, galaxy_light
        ):

            # Overwrite one value so intensity in each pixel is different
            grid_stack_0.sub[5] = np.array([2.0, 2.0])

            lp = galaxy_light.light_profiles[0]

            lp_sub_image = lp.intensities_from_grid(grid_stack_0.sub)

            # Perform sub gridding average manually
            lp_image_pixel_0 = (
                lp_sub_image[0] + lp_sub_image[1] + lp_sub_image[2] + lp_sub_image[3]
            ) / 4
            lp_image_pixel_1 = (
                lp_sub_image[4] + lp_sub_image[5] + lp_sub_image[6] + lp_sub_image[7]
            ) / 4

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_light], grid_stacks=[grid_stack_0]
            )

            assert (plane_stack.image_plane_images_1d[0][0] == lp_image_pixel_0).all()
            assert (plane_stack.image_plane_images_1d[0][1] == lp_image_pixel_1).all()
            assert (
                plane_stack.image_plane_images[0]
                == grid_stack_0.regular.scaled_array_2d_from_array_1d(
                    plane_stack.image_plane_images_1d[0]
                )
            ).all()

        def test__image_from_plane__same_as_galaxy_images(
            self, grid_stack_0, galaxy_light
        ):

            # Overwrite one value so intensity in each pixel is different
            grid_stack_0.sub[5] = np.array([2.0, 2.0])

            galaxy_image = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_0.sub, galaxies=[galaxy_light]
            )

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_light], grid_stacks=[grid_stack_0, grid_stack_0]
            )

            assert (plane_stack.image_plane_images_1d[0] == galaxy_image).all()
            assert (
                plane_stack.image_plane_images[0]
                == grid_stack_0.regular.scaled_array_2d_from_array_1d(
                    plane_stack.image_plane_images_1d[0]
                )
            ).all()

            assert (plane_stack.image_plane_images_1d[1] == galaxy_image).all()
            assert (
                plane_stack.image_plane_images[1]
                == grid_stack_0.regular.scaled_array_2d_from_array_1d(
                    plane_stack.image_plane_images_1d[1]
                )
            ).all()

        def test__image_plane_image_of_galaxies__use_multiple_grids__get_multiple_images(
            self, grid_stack_0, grid_stack_1
        ):

            # Overwrite one value so intensity in each pixel is different
            grid_stack_0.sub[5] = np.array([2.0, 2.0])

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

            plane_stack = pl_stack.PlaneStack(
                galaxies=[g0, g1], grid_stacks=[grid_stack_0, grid_stack_1]
            )

            g0_image_grid_0 = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_0.sub, galaxies=[g0]
            )
            g1_image_grid_0 = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_0.sub, galaxies=[g1]
            )

            assert plane_stack.image_plane_images_1d[0] == pytest.approx(
                g0_image_grid_0 + g1_image_grid_0, 1.0e-4
            )
            assert (
                plane_stack.image_plane_images[0]
                == grid_stack_0.regular.scaled_array_2d_from_array_1d(
                    plane_stack.image_plane_images_1d[0]
                )
            ).all()

            assert (
                plane_stack.image_plane_images_1d_of_galaxies[0][0] == g0_image_grid_0
            ).all()
            assert (
                plane_stack.image_plane_images_1d_of_galaxies[0][1] == g1_image_grid_0
            ).all()

            g0_image_grid_1 = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_1.sub, galaxies=[g0]
            )
            g1_image_grid_1 = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_1.sub, galaxies=[g1]
            )

            assert plane_stack.image_plane_images_1d[1] == pytest.approx(
                g0_image_grid_1 + g1_image_grid_1, 1.0e-4
            )
            assert (
                plane_stack.image_plane_images[1]
                == grid_stack_0.regular.scaled_array_2d_from_array_1d(
                    plane_stack.image_plane_images_1d[1]
                )
            ).all()

            assert (
                plane_stack.image_plane_images_1d_of_galaxies[1][0] == g0_image_grid_1
            ).all()
            assert (
                plane_stack.image_plane_images_1d_of_galaxies[1][1] == g1_image_grid_1
            ).all()

        def test__padded_grid_stack_in__image_plane_image_is_padded(
            self, padded_grid_stack, galaxy_light
        ):

            lp = galaxy_light.light_profiles[0]

            lp_sub_image = lp.intensities_from_grid(padded_grid_stack.sub)

            # Perform sub gridding average manually
            lp_image_pixel_0 = (
                lp_sub_image[0] + lp_sub_image[1] + lp_sub_image[2] + lp_sub_image[3]
            ) / 4
            lp_image_pixel_1 = (
                lp_sub_image[4] + lp_sub_image[5] + lp_sub_image[6] + lp_sub_image[7]
            ) / 4
            lp_image_pixel_2 = (
                lp_sub_image[8] + lp_sub_image[9] + lp_sub_image[10] + lp_sub_image[11]
            ) / 4
            lp_image_pixel_3 = (
                lp_sub_image[12]
                + lp_sub_image[13]
                + lp_sub_image[14]
                + lp_sub_image[15]
            ) / 4
            lp_image_pixel_4 = (
                lp_sub_image[16]
                + lp_sub_image[17]
                + lp_sub_image[18]
                + lp_sub_image[19]
            ) / 4
            lp_image_pixel_5 = (
                lp_sub_image[20]
                + lp_sub_image[21]
                + lp_sub_image[22]
                + lp_sub_image[23]
            ) / 4
            lp_image_pixel_6 = (
                lp_sub_image[24]
                + lp_sub_image[25]
                + lp_sub_image[26]
                + lp_sub_image[27]
            ) / 4
            lp_image_pixel_7 = (
                lp_sub_image[28]
                + lp_sub_image[29]
                + lp_sub_image[30]
                + lp_sub_image[31]
            ) / 4
            lp_image_pixel_8 = (
                lp_sub_image[32]
                + lp_sub_image[33]
                + lp_sub_image[34]
                + lp_sub_image[35]
            ) / 4
            lp_image_pixel_9 = (
                lp_sub_image[36]
                + lp_sub_image[37]
                + lp_sub_image[38]
                + lp_sub_image[39]
            ) / 4
            lp_image_pixel_10 = (
                lp_sub_image[40]
                + lp_sub_image[41]
                + lp_sub_image[42]
                + lp_sub_image[43]
            ) / 4
            lp_image_pixel_11 = (
                lp_sub_image[44]
                + lp_sub_image[45]
                + lp_sub_image[46]
                + lp_sub_image[47]
            ) / 4

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_light],
                grid_stacks=[padded_grid_stack, padded_grid_stack],
            )

            assert plane_stack.image_plane_images_for_simulation[0].shape == (3, 4)
            assert (
                plane_stack.image_plane_images_for_simulation[0][0, 0]
                == lp_image_pixel_0
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][0, 1]
                == lp_image_pixel_1
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][0, 2]
                == lp_image_pixel_2
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][0, 3]
                == lp_image_pixel_3
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][1, 0]
                == lp_image_pixel_4
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][1, 1]
                == lp_image_pixel_5
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][1, 2]
                == lp_image_pixel_6
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][1, 3]
                == lp_image_pixel_7
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][2, 0]
                == lp_image_pixel_8
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][2, 1]
                == lp_image_pixel_9
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][2, 2]
                == lp_image_pixel_10
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[0][2, 3]
                == lp_image_pixel_11
            ).all()

            assert plane_stack.image_plane_images_for_simulation[1].shape == (3, 4)
            assert (
                plane_stack.image_plane_images_for_simulation[1][0, 0]
                == lp_image_pixel_0
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][0, 1]
                == lp_image_pixel_1
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][0, 2]
                == lp_image_pixel_2
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][0, 3]
                == lp_image_pixel_3
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][1, 0]
                == lp_image_pixel_4
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][1, 1]
                == lp_image_pixel_5
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][1, 2]
                == lp_image_pixel_6
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][1, 3]
                == lp_image_pixel_7
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][2, 0]
                == lp_image_pixel_8
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][2, 1]
                == lp_image_pixel_9
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][2, 2]
                == lp_image_pixel_10
            ).all()
            assert (
                plane_stack.image_plane_images_for_simulation[1][2, 3]
                == lp_image_pixel_11
            ).all()

    class TestBlurringImages:
        def test__images_from_plane__same_as_their_light_profile(
            self, grid_stack_0, grid_stack_1, galaxy_light
        ):

            # Overwrite one value so intensity in each pixel is different
            grid_stack_0.blurring[1] = np.array([2.0, 2.0])

            lp = galaxy_light.light_profiles[0]

            lp_blurring_image = lp.intensities_from_grid(grid_stack_0.blurring)
            lp_blurring_image_1 = lp.intensities_from_grid(grid_stack_1.blurring)

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_light], grid_stacks=[grid_stack_0, grid_stack_1]
            )

            assert (
                plane_stack.image_plane_blurring_images_1d[0] == lp_blurring_image
            ).all()
            assert (
                plane_stack.image_plane_blurring_images_1d[1] == lp_blurring_image_1
            ).all()

        def test__same_as_above_but_for_galaxies(
            self, grid_stack_0, grid_stack_1, galaxy_light
        ):

            # Overwrite one value so intensity in each pixel is different
            grid_stack_0.blurring[1] = np.array([2.0, 2.0])

            galaxy_image = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_0.blurring, galaxies=[galaxy_light]
            )
            galaxy_image_1 = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_1.blurring, galaxies=[galaxy_light]
            )

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_light], grid_stacks=[grid_stack_0, grid_stack_1]
            )

            assert (plane_stack.image_plane_blurring_images_1d[0] == galaxy_image).all()
            assert (
                plane_stack.image_plane_blurring_images_1d[1] == galaxy_image_1
            ).all()

        def test__same_as_above__multiple_galaxies(self, grid_stack_0, grid_stack_1):

            # Overwrite one value so intensity in each pixel is different
            grid_stack_0.blurring[1] = np.array([2.0, 2.0])
            grid_stack_1.blurring[1] = np.array([2.0, 2.0])

            g0 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=1.0))
            g1 = g.Galaxy(light_profile=lp.EllipticalSersic(intensity=2.0))

            g0_image_grid_0 = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_0.blurring, galaxies=[g0]
            )
            g1_image_grid_0 = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_0.blurring, galaxies=[g1]
            )

            plane_stack = pl_stack.PlaneStack(
                galaxies=[g0, g1], grid_stacks=[grid_stack_0, grid_stack_1]
            )

            assert (
                plane_stack.image_plane_blurring_images_1d[0]
                == g0_image_grid_0 + g1_image_grid_0
            ).all()

            g0_image_grid_1 = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_1.blurring, galaxies=[g0]
            )
            g1_image_grid_1 = galaxy_util.intensities_of_galaxies_from_grid(
                grid_stack_1.blurring, galaxies=[g1]
            )

            plane_stack = pl_stack.PlaneStack(
                galaxies=[g0, g1], grid_stacks=[grid_stack_0, grid_stack_1]
            )

            assert (
                plane_stack.image_plane_blurring_images_1d[1]
                == g0_image_grid_1 + g1_image_grid_1
            ).all()

    class TestSurfaceDensity:
        def test__computes_x1_surface_density__uses_primary_grid_stack_0(
            self, grid_stack_0, grid_stack_1, galaxy_mass
        ):

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.convergence_from_grid(grid_stack_0.sub.unlensed_grid_1d)

            # Perform sub gridding average manually
            mp_image_pixel_0 = (
                mp_sub_image[0] + mp_sub_image[1] + mp_sub_image[2] + mp_sub_image[3]
            ) / 4
            mp_image_pixel_1 = (
                mp_sub_image[4] + mp_sub_image[5] + mp_sub_image[6] + mp_sub_image[7]
            ) / 4

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_mass], grid_stacks=[grid_stack_0, grid_stack_1]
            )

            assert (plane_stack.surface_density[1, 1] == mp_image_pixel_0).all()
            assert (plane_stack.surface_density[1, 2] == mp_image_pixel_1).all()

    class TestPotential:
        def test__computes_x1_potential__uses_primary_grid_stack_0(
            self, grid_stack_0, grid_stack_1, galaxy_mass
        ):

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.potential_from_grid(grid_stack_0.sub.unlensed_grid_1d)

            # Perform sub gridding average manually
            mp_image_pixel_0 = (
                mp_sub_image[0] + mp_sub_image[1] + mp_sub_image[2] + mp_sub_image[3]
            ) / 4
            mp_image_pixel_1 = (
                mp_sub_image[4] + mp_sub_image[5] + mp_sub_image[6] + mp_sub_image[7]
            ) / 4

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_mass], grid_stacks=[grid_stack_0, grid_stack_1]
            )

            assert (plane_stack.potential[1, 1] == mp_image_pixel_0).all()
            assert (plane_stack.potential[1, 2] == mp_image_pixel_1).all()

    class TestDeflections:
        def test__computes_x1_deflections__uses_primary_grid_stack_0(
            self, grid_stack_0, grid_stack_1, galaxy_mass
        ):

            mp = galaxy_mass.mass_profiles[0]

            mp_sub_image = mp.deflections_from_grid(grid_stack_0.sub.unlensed_grid_1d)

            # Perform sub gridding average manually
            mp_image_pixel_0x = (
                mp_sub_image[0, 0]
                + mp_sub_image[1, 0]
                + mp_sub_image[2, 0]
                + mp_sub_image[3, 0]
            ) / 4
            mp_image_pixel_1x = (
                mp_sub_image[4, 0]
                + mp_sub_image[5, 0]
                + mp_sub_image[6, 0]
                + mp_sub_image[7, 0]
            ) / 4
            mp_image_pixel_0y = (
                mp_sub_image[0, 1]
                + mp_sub_image[1, 1]
                + mp_sub_image[2, 1]
                + mp_sub_image[3, 1]
            ) / 4
            mp_image_pixel_1y = (
                mp_sub_image[4, 1]
                + mp_sub_image[5, 1]
                + mp_sub_image[6, 1]
                + mp_sub_image[7, 1]
            ) / 4

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_mass], grid_stacks=[grid_stack_0, grid_stack_1]
            )

            assert (plane_stack.deflections_1d[0, 0] == mp_image_pixel_0x).all()
            assert (plane_stack.deflections_1d[0, 1] == mp_image_pixel_0y).all()
            assert (plane_stack.deflections_1d[1, 0] == mp_image_pixel_1x).all()
            assert (plane_stack.deflections_1d[1, 1] == mp_image_pixel_1y).all()
            assert (
                plane_stack.deflections_y
                == grid_stack_0.regular.scaled_array_2d_from_array_1d(
                    plane_stack.deflections_1d[:, 0]
                )
            ).all()
            assert (
                plane_stack.deflections_x
                == grid_stack_0.regular.scaled_array_2d_from_array_1d(
                    plane_stack.deflections_1d[:, 1]
                )
            ).all()

    class TestMapper:
        def test__no_galaxies_with_pixelizations_in_plane__returns_none(
            self, grid_stack_0
        ):
            galaxy_no_pix = g.Galaxy()

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_no_pix],
                grid_stacks=[grid_stack_0],
                borders=[MockBorders()],
            )

            assert plane_stack.mapper is None

        def test__1_galaxy_in_plane__it_has_pixelization__returns_mapper(
            self, grid_stack_0
        ):
            galaxy_pix = g.Galaxy(
                pixelization=MockPixelization(value=1),
                regularization=MockRegularization(value=0),
            )

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_pix],
                grid_stacks=[grid_stack_0],
                borders=[MockBorders()],
            )

            assert plane_stack.mapper == 1

        def test__2_galaxies_in_plane__1_has_pixelization__extracts_reconstructor(
            self, grid_stack_0
        ):
            galaxy_pix = g.Galaxy(
                pixelization=MockPixelization(value=1),
                regularization=MockRegularization(value=0),
            )
            galaxy_no_pix = g.Galaxy()

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_no_pix, galaxy_pix],
                grid_stacks=[grid_stack_0],
                borders=[MockBorders()],
            )

            assert plane_stack.mapper == 1

        def test__plane_has_no_border__still_returns_mapper(self, grid_stack_0):
            galaxy_pix = g.Galaxy(
                pixelization=MockPixelization(value=1),
                regularization=MockRegularization(value=0),
            )
            galaxy_no_pix = g.Galaxy()

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_no_pix, galaxy_pix], grid_stacks=[grid_stack_0]
            )

            assert plane_stack.mapper == 1

        def test__2_galaxies_in_plane__both_have_pixelization__raises_error(
            self, grid_stack_0
        ):
            galaxy_pix_0 = g.Galaxy(
                pixelization=MockPixelization(value=1),
                regularization=MockRegularization(value=0),
            )
            galaxy_pix_1 = g.Galaxy(
                pixelization=MockPixelization(value=2),
                regularization=MockRegularization(value=0),
            )

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_pix_0, galaxy_pix_1],
                grid_stacks=[grid_stack_0],
                borders=[MockBorders()],
            )

            with pytest.raises(exc.PixelizationException):
                plane_stack.mapper

    class TestRegularization:
        def test__no_galaxies_with_pixelizations_in_plane__returns_none(
            self, grid_stack_0
        ):
            galaxy_no_pix = g.Galaxy()

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_no_pix],
                grid_stacks=[grid_stack_0],
                borders=[MockBorders()],
            )

            assert plane_stack.regularization is None

        def test__1_galaxy_in_plane__it_has_pixelization__returns_mapper(
            self, grid_stack_0
        ):
            galaxy_pix = g.Galaxy(
                pixelization=MockPixelization(value=1),
                regularization=MockRegularization(value=0),
            )

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_pix],
                grid_stacks=[grid_stack_0],
                borders=[MockBorders()],
            )

            assert plane_stack.regularization.value == 0

        def test__2_galaxies_in_plane__1_has_pixelization__extracts_reconstructor(
            self, grid_stack_0
        ):
            galaxy_pix = g.Galaxy(
                pixelization=MockPixelization(value=1),
                regularization=MockRegularization(value=0),
            )
            galaxy_no_pix = g.Galaxy()

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_no_pix, galaxy_pix],
                grid_stacks=[grid_stack_0],
                borders=[MockBorders()],
            )

            assert plane_stack.regularization.value == 0

        def test__2_galaxies_in_plane__both_have_pixelization__raises_error(
            self, grid_stack_0
        ):
            galaxy_pix_0 = g.Galaxy(
                pixelization=MockPixelization(value=1),
                regularization=MockRegularization(value=0),
            )
            galaxy_pix_1 = g.Galaxy(
                pixelization=MockPixelization(value=2),
                regularization=MockRegularization(value=0),
            )

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy_pix_0, galaxy_pix_1],
                grid_stacks=[grid_stack_0],
                borders=[MockBorders()],
            )

            with pytest.raises(exc.PixelizationException):
                plane_stack.regularization

    class TestPlaneImages:
        def test__x2_data_stacks__x2_plane_images(self, grid_stack_0, grid_stack_1):
            grid_stack_0.regular[1] = np.array([2.0, 2.0])

            galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))

            plane_stack = pl_stack.PlaneStack(
                galaxies=[galaxy],
                grid_stacks=[grid_stack_0, grid_stack_1],
                compute_deflections=False,
            )

            plane_image_from_func = lens_util.plane_image_of_galaxies_from_grid(
                shape=(3, 4), grid=grid_stack_0.regular, galaxies=[galaxy]
            )

            plane_image_from_func_1 = lens_util.plane_image_of_galaxies_from_grid(
                shape=(3, 4), grid=grid_stack_1.regular, galaxies=[galaxy]
            )

            assert (plane_image_from_func == plane_stack.plane_images[0]).all()
            assert (plane_image_from_func_1 == plane_stack.plane_images[1]).all()

        def test__ensure_index_of_plane_image_has_negative_arcseconds_at_start(
            self, grid_stack_0
        ):
            # The grid coordinates -2.0 -> 2.0 mean a plane of shape (5,5) has arc second coordinates running over
            # -1.6, -0.8, 0.0, 0.8, 1.6. The origin -1.6, -1.6 of the model_galaxy means its brighest pixel should be
            # index 0 of the 1D grid and (0,0) of the 2d plane datas_.

            msk = mask.Mask(array=np.full((5, 5), False), pixel_scale=1.0)

            grid_stack_0.regular = grids.Grid(
                np.array([[-2.0, -2.0], [2.0, 2.0]]), mask=msk
            )

            g0 = g.Galaxy(
                light_profile=lp.EllipticalSersic(centre=(1.6, -1.6), intensity=1.0)
            )
            plane_stack = pl_stack.PlaneStack(galaxies=[g0], grid_stacks=[grid_stack_0])

            assert plane_stack.plane_images[0].shape == (5, 5)
            assert np.unravel_index(
                plane_stack.plane_images[0].argmax(), plane_stack.plane_images[0].shape
            ) == (0, 0)

            g0 = g.Galaxy(
                light_profile=lp.EllipticalSersic(centre=(1.6, 1.6), intensity=1.0)
            )
            plane_stack = pl_stack.PlaneStack(galaxies=[g0], grid_stacks=[grid_stack_0])
            assert np.unravel_index(
                plane_stack.plane_images[0].argmax(), plane_stack.plane_images[0].shape
            ) == (0, 4)

            g0 = g.Galaxy(
                light_profile=lp.EllipticalSersic(centre=(-1.6, -1.6), intensity=1.0)
            )
            plane_stack = pl_stack.PlaneStack(galaxies=[g0], grid_stacks=[grid_stack_0])
            assert np.unravel_index(
                plane_stack.plane_images[0].argmax(), plane_stack.plane_images[0].shape
            ) == (4, 0)

            g0 = g.Galaxy(
                light_profile=lp.EllipticalSersic(centre=(-1.6, 1.6), intensity=1.0)
            )
            plane_stack = pl_stack.PlaneStack(galaxies=[g0], grid_stacks=[grid_stack_0])
            assert np.unravel_index(
                plane_stack.plane_images[0].argmax(), plane_stack.plane_images[0].shape
            ) == (4, 4)
